package com.unitesk.requality.core;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.UUID;

import org.json.JSONException;

public class FSStorage extends TreeStorage {
    private static final String ROOT_NODE_FILENAME = "root.json";
    private Map<File, UUID> fileToNode = new HashMap<File, UUID>();
    private String path;

    public FSStorage(String path) {
        if(! path.endsWith(File.separator)) {
            path=path+File.separator;
        }
        this.path = path;
    }

    private UUID getUUId(File file) {
        return new NodeMapper(readFile(file)).getUUId();
    }

    public File getFile(UUID uuid) {
        for (File r : fileToNode.keySet()) {
            if (fileToNode.get(r).equals(uuid))
                return r;
        }
        return null;
    }

    private String readFile(File file) {
        try {
            StringBuilder jsonsb = new StringBuilder();
            Scanner s = new Scanner(new FileInputStream(file));
            while (s.hasNextLine()) {
                jsonsb.append(s.nextLine().trim());
            }
            return jsonsb.toString();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public UUID getRootNode() {
        File file = new File(path + ROOT_NODE_FILENAME);
        if (file.exists()) {
            UUID uuid = getUUId(file);
            fileToNode.put(file, uuid);
            return uuid;
        }
        return null;
    }

    public File getResourceFolder(UUID uuid) {
        String id = getNodeId(uuid);
        File resFolder = new File(getFile(uuid).getParent() + File.separator
                + id + "_resources");
        return resFolder;
    }

    @Override
    public String getNodeId(UUID uuid) {
        if (uuid.equals(getRootNode())) {
            return "/";
        }
        if (getFile(uuid) == null) {
            return null; // node not found
        }
        String jext = "json";
        String fileName = getFile(uuid).getName();
        return fileName.substring(0, fileName.length() - (jext.length() + 1));
    }

    @Override
    public UUID getParent(UUID uuid) {
        if (getFile(uuid) == null) {
            return null; // node not found
        }
        return getParent(getFile(uuid));
    }

    public UUID getParent(File file) {
        if (file.getAbsoluteFile().equals(path)) {
            return null;
        }

        String folderName = file.getParentFile().getName();
        String parentId = folderName.substring(0, folderName.length());

        File parentFolder = file.getParentFile().getParentFile();
        File parentResource = new File(parentFolder.getAbsoluteFile()
                + File.separator + parentId + ".json");

        if (!parentResource.exists()) {
            return null;
        }
        return getUUId(parentResource);
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        Collection<UUID> result = new HashSet<UUID>();
        if (getFile(uuid) == null) {
            return result; // node not found
        }

        String nodeId = getNodeId(uuid);
        if(nodeId.equals("/")) {
            nodeId="root";
        }
        File nodeFolder = getFile(uuid).getParentFile();
        File childrenFolder = new File(nodeFolder.getAbsoluteFile()
                + File.separator + nodeId);

        if (!childrenFolder.exists()) {
            return result;
        }

        for (File res : childrenFolder.listFiles()) {
            if (res.isFile() && res.getName().endsWith(".json")) {
                UUID childUUID = getUUId(res);
                result.add(getUUId(res));
                fileToNode.put(res, childUUID);
            }
        }
        return result;
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        try {
            File file;
            if (parentUUId == null) {
                file = new File(path + ROOT_NODE_FILENAME);
            } else {
                String parentId = getNodeId(parentUUId);
                if (parentId.equals("/")) {
                    parentId = "root";
                }
                File childFolder = new File(getFile(parentUUId).getParentFile()
                        .getAbsoluteFile() + File.separator + parentId);

                if (!childFolder.exists()) {
                    childFolder.mkdirs();
                }
                file = new File(childFolder.getAbsolutePath()+ File.separator + id + ".json");
            }

            if(! file.getParentFile().exists()) {
                file.getParentFile().mkdirs();
            }
            if(! file.exists()) {
                file.createNewFile();
            }
            FileOutputStream out = new FileOutputStream(file);
            Writer w = new OutputStreamWriter(out, "UTF-8");
            String jsonString = NodeMapper.toJSON(uuid, attrs).toString(4);
            w.append(jsonString);
            w.flush();
            w.close();

            // update model
            fileToNode.put(file, uuid);
            deltaTarget.created(uuid);

        } catch (UnsupportedEncodingException e) {
            e.printStackTrace();
        } catch (JSONException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public void remove(UUID uuid) {
        // TODO: implement me
//        String id = getNodeId(uuid);
//        File res = getFile(uuid);
//        try {
//            new File(res.getParentFile()+id);
//            res.delete(false, null);
//        } catch (CoreException e) {
//            e.printStackTrace();
//        }
    }

    @Override
    public void changeNodeId(UUID node, String id) {
        // TODO Auto-generated method stub

    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        NodeMapper m = new NodeMapper(readFile(getFile(uuid)));
        return m.getAttributes();
    }

    @Override
    public void saveAttributes(UUID node, Collection<Attribute> attributes) {
        // TODO Auto-generated method stub

    }

    @Override
    public void moveNodeTo(UUID node, UUID parent) {
        // TODO Auto-generated method stub

    }

    @Override
    public boolean canLoad(UUID uuid) {
        // TODO Auto-generated method stub
        return false;
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        return getResourceFolder(uuId).getAbsolutePath();
    }
}
