package com.unitesk.requality.merge;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.unitesk.requality.merge.Location.InvalidLocationException;

public class MergeProcessor {

    private static final MergeProcessor instance = new MergeProcessor();
    private List<String> idsNotFound;
    private List<String> idsFound;
    private List<Location> locations;

    private MergeProcessor() {
	idsNotFound = new ArrayList<String>();
	idsFound = new ArrayList<String>();
	locations = new ArrayList<Location>();
    }

    public static MergeProcessor getInstance() {
	return instance;
    }

    public Document merge(Document v1, Document v2) {
	if (!idsNotFound.isEmpty())
	    idsNotFound.clear();
	if (!idsFound.isEmpty())
	    idsFound.clear();
	if (!locations.isEmpty())
	    locations.clear();
	Tree t1 = new Tree(v1);
	Tree t2 = new Tree(v2);
	locations = lookupLocations(t1);
	locations = traceLocations(locations, t2);
	Map<Node, List<Location>> locationMapping = mapLocations(locations);
	Map<Node, Fragment> fragmentMapping = makeFragments(locationMapping);
	Iterator<Node> i = fragmentMapping.keySet().iterator();
	while (i.hasNext()) {
	    Node anchorNode = i.next();
	    mergeFragment(fragmentMapping.get(anchorNode), anchorNode);
	    t2.updateTree();
	}
	return v2;
    }

    private Map<Node, Fragment> makeFragments(
	    Map<Node, List<Location>> locationMapping) {
	Map<Node, Fragment> mapping = new HashMap<Node, Fragment>();
	Iterator<Node> i = locationMapping.keySet().iterator();
	while (i.hasNext()) {
	    Node anchorNode = i.next();
	    List<Location> locations = locationMapping.get(anchorNode);
	    List<Node> relatedNodes = new ArrayList<Node>();
	    for (Location loc : locations) {
		List<Node> locNodes = loc.getRelatedNodes();
		for (Node x : locNodes)
		    if (!relatedNodes.contains(x))
			relatedNodes.add(x);
	    }
	    Fragment f = new Fragment();
	    for (Node x : relatedNodes)
		f.attachNode(x, true);
	    mapping.put(anchorNode, f);
	}
	return mapping;
    }

    private void mergeFragment(Fragment f, Node anchorNode) {
	org.w3c.dom.Node anchorDomNode = anchorNode.getDomNode();
	org.w3c.dom.Node fRoot = f.getRoot().getDomNode();
	fRoot = anchorDomNode.getOwnerDocument().importNode(fRoot, true);
	anchorDomNode.getParentNode().replaceChild(fRoot, anchorDomNode);
    }

    private Map<Node, List<Location>> mapLocations(List<Location> locations) {
	Map<Node, List<Location>> mapping = new HashMap<Node, List<Location>>();
	checkIds(locations);
	for (Location loc : locations) {
	    String id = loc.getLocationId();
	    Node anchorNode = loc.getAnchorNode();
	    if (anchorNode != null && !idsNotFound.contains(id)
		    && idsFound.contains(id)) {
		if (mapping.containsKey(anchorNode))
		    mapping.get(anchorNode).add(loc);
		else {
		    List<Location> list = new ArrayList<Location>();
		    list.add(loc);
		    mapping.put(anchorNode, list);
		}

	    }
	}
	return mapping;
    }

    private void checkIds(List<Location> locations) {
	for (Location loc : locations) {
	    Node anchorNode = loc.getAnchorNode();
	    String id = loc.getLocationId();
	    if (anchorNode == null) {
		if (!idsNotFound.contains(id))
		    idsNotFound.add(id);
		if (idsFound.contains(id))
		    idsFound.remove(id);
	    } else if (!idsNotFound.contains(id) && !idsFound.contains(id))
		idsFound.add(id);
	}
    }

    private List<Location> traceLocations(List<Location> locations, Tree t) {
	List<Location> newLocations = new ArrayList<Location>(locations);
	Collections.copy(newLocations, locations);
	for (Location loc : newLocations) {
	    Node newAnchorNode = traceLocation(loc, t);
	    loc.setAnchorNode(newAnchorNode);
	}
	return newLocations;
    }

    private Node traceLocation(Location loc, Tree t) {
	System.out.println("Tracing " + loc);
	NodeFilter parentFilter = new NodeFilter(loc.getSpanNode().getParent(),
		true, true, false, false);
	Node spanParent = t.nearestMatch(loc.getSpanNode().getParent(),
		parentFilter);
	System.out.println("Anchor node parent: " + spanParent);
	if (spanParent != null
		&& ((hasUpdatedLocations(loc) && isCommonParent(loc, spanParent)) || !hasUpdatedLocations(loc))
		&& spanParent.isEqual(loc.getSpanNode().getParent())) {
	    NodeFilter spanFilter = new NodeFilter(loc.getAnchorNode(), true,
		    false, false, true);
	    Node spanNearestMatch = t.nearestMatch(loc.getSpanNode(),
		    spanParent, spanFilter);
	    System.out.println("Anchor node match(by span): "
		    + spanNearestMatch);
	    NodeFilter textFilter = spanFilter;
	    Node textNearestMatch = t.nearestMatch(loc.getAnchorNode(),
		    spanParent, textFilter);
	    System.out.println("Anchor node match(by text): "
		    + textNearestMatch);
	    if (spanNearestMatch != null
		    && textNearestMatch != null
		    && spanNearestMatch.equals(textNearestMatch)
		    && textNearestMatch.getDomNode().getNodeType() == org.w3c.dom.Node.TEXT_NODE) {
		textNearestMatch.getContext().setWeight(100);
		String nodeValue = textNearestMatch.getDomNode().getNodeValue()
			.trim();
		if (nodeValue.contains(loc.getAnchorText().trim()))
		    return textNearestMatch;
	    }
	}
	System.out.println("Not found " + loc);
	return null;
    }

    public void merge(Reader v1, Reader v2, Writer output) {
	try {
	    DocumentBuilderFactory factory = DocumentBuilderFactory
		    .newInstance();
	    DocumentBuilder builder = factory.newDocumentBuilder();
	    Document doc1 = builder.parse(new InputSource(v1));
	    Document doc2 = builder.parse(new InputSource(v2));
	    try {
		Document result = merge(doc1, doc2);
		doc2 = result;
	    } catch(NullPointerException e){
		e.printStackTrace();
		throw e;
	    } finally {
		writeXmlFile(doc2, output);
	    }
	} catch (ParserConfigurationException e) {
	    e.printStackTrace();
	} catch (SAXException e) {
	    e.printStackTrace();
	} catch (IOException e) {
	    e.printStackTrace();
	} 
    }

    private List<Location> lookupLocations(Tree t) {
	List<Location> locations = new ArrayList<Location>();
	for (Node x : t.getTraversalSequence()) {
	    if (Location.isLocationNode(x))
		try {
		    locations.add(new Location(x));
		} catch (InvalidLocationException e) {
		    e.printStackTrace();
		}
	}
	return locations;
    }

    private static void writeXmlFile(Document doc, Writer output) {
	try {
	    Source source = new DOMSource(doc);

	    Result result = new StreamResult(output);

	    Transformer xformer = TransformerFactory.newInstance()
		    .newTransformer();
	    xformer.transform(source, result);
	} catch (TransformerConfigurationException e) {
	} catch (TransformerException e) {
	}
    }

    private boolean hasUpdatedLocations(Location refLoc) {
	for (Location loc : getLocationsById(refLoc.getLocationId())) {
	    if (loc.equals(refLoc))
		continue;
	    if (loc.wasUpdated())
		return true;
	}
	return false;
    }

    private List<Location> getLocationsById(String id) {
	List<Location> sameIdLocations = new ArrayList<Location>();
	for (Location loc : locations)
	    if (loc.getLocationId().equals(id))
		sameIdLocations.add(loc);
	return sameIdLocations;
    }

    private boolean isCommonParent(Location refLocation, Node parent) {
	for (Location loc : getLocationsById(refLocation.getLocationId())) {
	    if (loc.equals(refLocation))
		continue;
	    Node currAnchorNode = loc.getAnchorNode();
	    if (currAnchorNode != null
		    && currAnchorNode.getDomNode().getOwnerDocument()
			    .isSameNode(parent.getDomNode().getOwnerDocument())) {
		Node currParent = currAnchorNode.getParent();
		Node parentsParent = parent.getParent();
		if (currParent != null
			&& (currParent.getDomNode().isSameNode(
				parent.getDomNode()) || (parentsParent != null && parentsParent
				.getDomNode().isSameNode(
					currParent.getDomNode()))))
		    return true;
	    }
	}
	return false;
    }

    public List<String> getMissingIds() {
	return this.idsNotFound;
    }

    public List<String> getMergedIds() {
	return this.idsFound;
    }
}
