package com.unitesk.requality.tests.core;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.UUID;

import org.junit.Test;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.TreeStorage;

abstract public class StorageTest {

    abstract protected TreeStorage createStorage();

    @Test
    public void testGetRootNode() {
        TreeNode root = getDB().getRootNode();
        assertNotNull(root);
        assertEquals(root.getUUId(), getDB().getStorage().getRootNode());
        assertTrue(getDB().getStorage().canLoad(root.getUUId()));
        assertNull(getDB().getStorage().getParent(root.getUUId()));
    }

    abstract protected TreeDB getDB();

    @Test
    public void getParent() {
        TreeNode root = getDB().getRootNode();
        assertNull(getDB().getStorage().getParent(root.getUUId()));
        TreeNode a = root.createChild("a");
        assertEquals(root.getUUId(), getDB().getStorage().getParent(a.getUUId()));
        // TODO: increas coverage
    }

    @Test
    public void getChildren() {
        getDB().getRootNode();
        UUID rootUUID = getDB().getStorage().getRootNode();
        assertEquals(0, getDB().getStorage().getChildren(rootUUID).size());
        TreeNode a = getDB().getRootNode().createChild("a");
        assertEquals(1, getDB().getStorage().getChildren(rootUUID).size());
        assertTrue(getDB().getStorage().getChildren(rootUUID).contains(a.getUUId()));
    }

    @Test
    public void createNode() {
        UUID rootUUID = getDB().getRootNode().getUUId();
        UUID uuid = UUID.randomUUID();
        getDB().getStorage().createNode(rootUUID, uuid , "Node", "a");
        assertEquals(1, getDB().getStorage().getChildren(rootUUID).size());
        assertEquals("Node", getDB().getStorage().getNodeType(uuid));
        assertEquals(rootUUID, getDB().getStorage().getParent(uuid));
        assertEquals("a", getDB().getStorage().getNodeId(uuid));
        assertTrue(getDB().getStorage().canLoad(uuid));

    }

    @Test
    public void remove() {
        UUID rootUUID = getDB().getRootNode().getUUId();
        UUID uuid = UUID.randomUUID();
        getDB().getStorage().createNode(rootUUID, uuid , "Node", "a");
        getDB().getStorage().remove(uuid);
        assertEquals(0, getDB().getStorage().getChildren(rootUUID).size());
        assertFalse(getDB().getStorage().canLoad(uuid));
    }

    @Test
    public void getNodeId() {
        UUID rootUUID = getDB().getRootNode().getUUId();
        UUID aUUID = getDB().getRootNode().createChild("a").getUUId();
        assertEquals("/", getDB().getStorage().getNodeId(rootUUID));
        assertEquals("a", getDB().getStorage().getNodeId(aUUID));

    }

    @Test
    public void changeNodeId() {
        UUID aUUID = getDB().getRootNode().createChild("a").getUUId();
        getDB().getStorage().changeNodeId(aUUID, "b");
        assertEquals("b", getDB().getNode("/b").getId());
        assertEquals("b", getDB().getStorage().getNodeId(aUUID));
    }

    @Test
    public void getAttributes() {
        TreeNode a = getDB().getRootNode().createChild("a");
        a.putAttribute(new Attribute(AttributeType.STRING, "v", "1"));
        a.saveAttributes();
        Collection<Attribute> attrs = getDB().getStorage().getAttributes(a.getUUId());
        assertEquals(1+1, attrs.size());
        // TODO: (test) check attributes
    }

    @Test
    public void saveAttributes() {
        TreeNode a = getDB().getRootNode().createChild("a");
        a.putAttribute(new Attribute(AttributeType.STRING, "v", "1"));
        assertEquals(0+1, getDB().getStorage().getAttributes(a.getUUId()).size());
        a.saveAttributes();
        assertEquals(1+1, getDB().getStorage().getAttributes(a.getUUId()).size());
    }

    @Test
    public void leafMoveNodeTo() {
        TreeNode root = getDB().getRootNode();
        TreeNode a = root.createChild("a");
        TreeNode aa = root.createChild("aa");
        aa.setParent(a);
        assertEquals(a, aa.getParent());
        assertEquals(aa, a.getChild(0));
        assertEquals(1, root.sizeChildren());
        assertEquals(1, a.sizeChildren());
        aa.setParent(root);
        assertEquals(root, aa.getParent());
        assertEquals(2, root.sizeChildren());
        assertEquals(a, root.findChild(a.getId()));
        assertEquals(0, a.sizeChildren());
    }

    @Test
    public void subtreeMoveNodeTo() {
        TreeNode root = getDB().getRootNode();
        TreeNode a = root.createChild("a");
        TreeNode aa = root.createChild("aa");
        TreeNode aaa = aa.createChild("aaa");
        aa.setParent(a);
        // TODO: check tree structure
        assertEquals(1, root.sizeChildren());
        assertEquals(a, root.findChild("a"));
        assertEquals(1, a.sizeChildren());
        assertEquals(aa, a.findChild("aa"));
        assertEquals(1, aa.sizeChildren());
        assertEquals(aaa, aa.findChild("aaa"));
    }

    @Test
    public void canLoad() {
        assertTrue(getDB().getStorage().canLoad(getDB().getRootNode().getUUId()));
        assertFalse(getDB().getStorage().canLoad(UUID.fromString("00000000-0000-0000-0000-000000000000")));
    }

}
